/**************************************************
opengate Mac addr auth program

 module to control ip addresss cache 

  As ip address check by database is time consuming procedure,
  the recently checked addresses are cached and skiped.
  Implemented with HashTable and Queue.
  HashTable has recentry checked ipAddress and checked time.
  If ip is included in table and time is new, ignore checking.
  Queue has ipAddresses odrered by checked time.
  If an old item is found in table, elder items are removed from table.
  The queue controls the remove sequence.


Copyright (C) 2011 Opengate Project Team
Written by Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/
#include "opengatemd.h"

int Initqueue(void);
int Enqueue(unsigned char* addrRaw, int addrLen);
int Dequeue(unsigned char* addrRaw, int* pAddrLen);
int Listqueue(void);
void Freequeue(void);

/* Queue to store IpAddress */
struct queueNode{
  int addrLen;
  unsigned char addrRaw[IPV6ADDRLN];
  struct queueNode *next;
};
static struct queueNode* queueTail=NULL;
static struct queueNode* queueHead=NULL;

/* HashTable to store IpAddress and Time */
static DB* hashDb;

/* Cache Timeout(seconds) = DB checking interval */
static int cacheTimeout;


/****************************************
initialize IpAddress Cache
****************************************/
void initIpCache(void) {

  DBT hashKey;
  DBT hashVal;
  int timeNull=0;
  unsigned char zeroAddrRaw[IPV6ADDRLN];
  
  /* prepare binary zero address */
  bzero(zeroAddrRaw, IPV6ADDRLN);

  /* prepare hash table */
  if((hashDb = dbopen(NULL, O_CREAT | O_RDWR, 0644, DB_HASH, NULL)) == NULL) {
    err_msg("ERR at %s#%d: fail to open hash table",__FILE__,__LINE__);
    terminateProg(0);
  }

  /* ingoring addresses(0.0.0.0 & 0::0) is registered to hash table */
  hashKey.data = zeroAddrRaw;
  hashKey.size = IPV4ADDRLN;
  hashVal.data = &timeNull;
  hashVal.size = sizeof(int);    
  if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
    err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
  }
  hashKey.data = zeroAddrRaw;
  hashKey.size = IPV6ADDRLN;
  hashVal.data = &timeNull;
  hashVal.size = sizeof(int);    
  if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
    err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
  }

  /* prepare queue */
  Initqueue();

  /* set timeout parameter */
  cacheTimeout=atoi(GetConfValue("CacheTimeout"));
}

/*********************************************
initialize IpAddress Queue
Queue
 HeadNode - DataNode - DataNode - TailNode
 (dummy)                           (dummy)
  ^queueHead                        ^queueTail
*********************************************/
int initqueue(void){

  unsigned char addrRaw[IPV6ADDRLN];
  int addrLen;

  /* if not exist, prepare head and tail */
  if(queueHead==NULL){
    queueHead=(struct queueNode*)malloc(sizeof(struct queueNode));
    if(queueHead==NULL){
      err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
      terminateProg(0);
    }
    queueTail=(struct queueNode*)malloc(sizeof(struct queueNode));
    if(queueTail==NULL){
      err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
      terminateProg(0);
    }
    bzero(queueHead->addrRaw, IPV6ADDRLN);
    queueHead->addrLen=0;
    bzero(queueTail->addrRaw, IPV6ADDRLN);
    queueTail->addrLen=0;
    queueHead->next=queueTail;
    queueTail->next=NULL;
  }
  
  /* if exist, reset all */
  else{
    while(Dequeue(addrRaw,&addrLen))
      ;
  }
  return TRUE;
}

/****************************************
Add data to the tail of Queue
 input=addr
****************************************/
int enqueue(unsigned char* addrRaw, int addrLen){
  struct queueNode *newNode;

  /* if not prepared, error */
  if(queueHead==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;
  }

  /* add item after the tail and set it as new tail*/
  newNode=(struct queueNode*)malloc(sizeof(struct queueNode));
  if(newNode==NULL){
    err_msg("ERR at %s#%d: fail to malloc",__FILE__,__LINE__);
    terminateProg(0);
  }
  memcpy(queueTail->addrRaw, addrRaw, addrLen);
  queueTail->addrLen=addrLen;
  queueTail->next=newNode;
  queueTail=newNode;
  bzero(queueTail->addrRaw,IPV6ADDRLN);
  queueTail->addrLen=0;
  queueTail->next=NULL;
  return TRUE;
}

/****************************************
Get and remove address data from the head of Queue
 output=addr
****************************************/
int dequeue(unsigned char* addrRaw, int* pAddrLen){

  /* set null string as default */
  bzero(addrRaw, IPV6ADDRLN);

  /* if not prepared, error */
  if(queueHead==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;
  }
  else if(queueHead->next==NULL){
    err_msg("ERR at %s#%d: queue not init",__FILE__,__LINE__);
    return FALSE;  
  }

  /* if no data, return false */
  else if(queueHead->next==queueTail){
    return FALSE;
  }

  /* get item from the head */
  else {
    struct queueNode *temp;
    temp=queueHead->next;
    queueHead->next=temp->next;
    memcpy(addrRaw, temp->addrRaw, temp->addrLen);
    *pAddrLen=temp->addrLen;
    free(temp);
  }
  return TRUE;
}

/****************************************
Listing IpAddress Queue (for debugging)
****************************************/
int listqueue(void){

  struct queueNode *temp;
  int i;

  if(queueHead==NULL) return FALSE;
  temp=queueHead->next;
  while(temp->next!=NULL){
    for(i=0;i<temp->addrLen;i++){
      printf("%x", temp->addrRaw[i]);
    }
    temp=temp->next;
  }
  return TRUE;
}


/****************************************
Is the IpAddress checked recently or not
 input=ipAddress return TRUE if checked recently
****************************************/
int isRecentlyCheckedIp(unsigned char* ipAddressRaw, int addrLen){

  int timeNow;
  unsigned char storedAddrRaw[IPV6ADDRLN];
  int storedAddrLen;
  int ret;
  int* pTime;
  DBT hashKey;
  DBT hashVal;

  /* get present time */
  timeNow=time(NULL);

  /* get item matched to the ip from hash table */
  hashKey.data = ipAddressRaw;
  hashKey.size = addrLen;
  memset(&hashVal, 0, sizeof(DBT));
  ret=hashDb->get(hashDb, &hashKey, &hashVal, 0);

  /* get success */
  if(ret==0){

    /* if data is null, return not checked */
    pTime=(int*)(hashVal.data);
    if(pTime==NULL) return FALSE;

    /* if ignoring addresses(eg, 0.0.0.0, ::), return true */
    if(*pTime==0) return TRUE;

    /* if recently checked data, return true */
    if( (timeNow-*pTime) < cacheTimeout ) return TRUE;

    /* if timeover, elder items are removed from queue and hashTable */
    /* get data from queue head and inspect the address */
    /* until present address is found. */
    while(Dequeue(storedAddrRaw, &storedAddrLen)){
      hashKey.data=storedAddrRaw;
      hashKey.size = storedAddrLen;
      hashDb->del(hashDb, &hashKey, 0);
      if(memcmp(ipAddressRaw,storedAddrRaw,storedAddrLen)==0)break;
    }

    /* insert update item to queue and hashTable */
    Enqueue(ipAddressRaw, storedAddrLen);
    hashVal.data = &timeNow;
    hashVal.size = sizeof(int);    
    if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
      err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
      terminateProg(0);
    }

    /* and return false */
    return FALSE;
  }

  /* if getting from hash fails, insert address and return false */
  if(ret==1){

    /* insert to hash table */
    hashVal.data = &timeNow;
    hashVal.size = sizeof(int);    
    if(hashDb->put(hashDb, &hashKey, &hashVal, 0) == -1) {
      err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
      terminateProg(0);
    }

    /* insert to queue */
    Enqueue(ipAddressRaw, addrLen);
    return FALSE;
  }

  /* else error exit */
  err_msg("ERR at %s#%d: fail to get hash table item",__FILE__,__LINE__);
  return FALSE;
}

/****************************************
memory free for IpAddress Cache
****************************************/
void freeIpCache(void) {

  hashDb->close(hashDb);
  Freequeue();
}

/****************************************
memory free for IpAddress Queue
****************************************/
void freequeue(void){
  unsigned char addrRaw[ADDRMAXLN];
  int addrLen;
  while(Dequeue(addrRaw,&addrLen));
  free(queueHead);
  free(queueTail);
}

/****************************************************
 routines for debugging putput
 ***************************************************/

void InitIpCache(void) {
  if(debug>1) err_msg("DEBUG:=>initIpCache( )");
  initIpCache();
  if(debug>1) err_msg("DEBUG:<=initIpCache( )");
}
int Initqueue(void){
  int ret;
  if(debug>1) err_msg("DEBUG:=>initqueue( )");
  ret = initqueue();
  if(debug>1) err_msg("DEBUG:(%d)<=initqueue( )",ret);
  return ret;

}
int Enqueue(unsigned char* addrRaw, int addrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>enqueue(%x,%d)", addrRaw[0],addrLen);
  ret = enqueue(addrRaw,addrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=enqueue( )",ret);
  return ret;
}

int Dequeue(unsigned char* addrRaw, int* pAddrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>dequeue( )");
  ret = dequeue(addrRaw,pAddrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=dequque(%x,%d)",ret, addrRaw[0],*pAddrLen);
  return ret;

}
int Listqueue(void){
  int ret;
  if(debug>1) err_msg("DEBUG:=>listQueue( )");
  ret = listqueue();
  if(debug>1) err_msg("DEBUG:(%d)<=listQueue( )",ret);
  return ret;

}

int IsRecentlyCheckedIp(unsigned char* ipAddressRaw, int addrLen){
  int ret;
  if(debug>1) err_msg("DEBUG:=>isRecentlyCheckedIp(%x,%d)", ipAddressRaw[0],addrLen);
  ret = isRecentlyCheckedIp(ipAddressRaw, addrLen);
  if(debug>1) err_msg("DEBUG:(%d)<=isRecentlyCheckedIp( )",ret);
  return ret;
}

void FreeIpCache(void) {
  if(debug>1) err_msg("DEBUG:=>freeIpCache()");
  freeIpCache();
  if(debug>1) err_msg("DEBUG:<=freeIpCache()");


}

void Freequeue(void){
  if(debug>1) err_msg("DEBUG:=>freeQueue()");
  freequeue();
  if(debug>1) err_msg("DEBUG:<=freequeue()");

}
/**************** test main *************************
int testmain(){

  char addr[ADDRMAXLN];

  initIpCache();

  while(1){

    printf("Enter addr=");
    scanf("%s", addr);

    if(isRecentlyCheckedIp(addr)){
      printf("new\n");
    }else{
      printf("old\n");
    }
    listqueue();
  }
  return 0;
}
*************************************************/

